from __future__ import print_function
import numpy as np
import math
import matplotlib
matplotlib.use('agg')
from scipy import interpolate
import matplotlib.pyplot as plt
import mpld3
from . import ugtm_landscape
import json
from mpld3 import _display


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


_display.NumpyEncoder = NumpyEncoder


def plot_pdf(coordinates, labels=None, title="", output="output",
             discrete=False, pointsize=1.0, alpha=0.3,cname="Spectral_r"):
    if labels is None:
        colvec = "black"
    elif discrete:
        uniqClasses, label_numeric = np.unique(labels, return_inverse=True)
        colvec = np.squeeze(label_numeric)
    elif not discrete:
        label_numeric = labels
        colvec = np.squeeze(label_numeric)
    fig, ax = plt.subplots()
    ax.grid(color='white', linestyle='solid')
    ax.set_axisbelow(True)
    ax.set_title(title, size=30)
    scatter = ax.scatter(coordinates[:, 0].tolist(),
                         coordinates[:, 1].tolist(),
                         c=colvec, s=20*pointsize, alpha=alpha,
                         cmap=plt.get_cmap(cname),
                         edgecolor='black')
    plt.xlim(-1.1, 1.1)
    plt.ylim(-1.1, 1.1)
    if not discrete and (labels is not None):
        plt.colorbar(scatter)
    plt.axis('tight')
    plt.xticks([]), plt.yticks([])
    fig.savefig(output+".pdf", format='pdf', dpi=500)
    plt.close(fig)
    print("\nWrote pdf to disk: %s\n" % (output+".pdf"))


def plotMultiPanelGTM(optimizedModel, labels, output="output", discrete=False,
                      pointsize=1.0, alpha=0.3, do_interpolate=True,
                      cname="Spectral_r"):
    if labels is None:
        colvec = "black"
    elif labels is not None and discrete:
        uniqClasses, label_numeric = np.unique(labels, return_inverse=True)
        colvec = np.squeeze(label_numeric)
    elif not discrete:
        label_numeric = labels
        colvec = np.squeeze(label_numeric)
    fig = plt.figure(figsize=(10, 10))
    means = optimizedModel.matMeans
    modes = optimizedModel.matModes
    # plot1: GTM means visualization
    ax = fig.add_subplot(221)
    ax.scatter(means[:, 0], means[:, 1], c=colvec,
               cmap=plt.get_cmap(cname), s=20*pointsize, alpha=alpha,
               edgecolor="black")
    plt.axis('tight')
    plt.xticks([])
    plt.yticks([])
    plt.title('Means')
    # plot2: GTM modes visualization
    ax2 = fig.add_subplot(222)
    ax2.scatter(modes[:, 0], modes[:, 1], c=colvec,
                cmap=plt.get_cmap(cname), s=20*pointsize, alpha=alpha,
                edgecolor="black")
    plt.axis('tight')
    plt.xticks([]), plt.yticks([])
    plt.title('Modes')
    # plot3: GTM landscape visualization
    ax3 = fig.add_subplot(223)
    # if it's label data, the landscape is a class map;
    # otherwise, it is a continuous landscape
    if labels is not None and discrete:
        plotClassMapNoPoints(optimizedModel, label_numeric,
                             do_interpolate=do_interpolate, cname=cname)
    elif labels is not None:
        plotLandscapeNoPoints(optimizedModel, label_numeric,
                              do_interpolate=do_interpolate, cname=cname)
    ax3.scatter(means[:, 0], means[:, 1], c=colvec,
                cmap=plt.get_cmap(cname),
                s=20*pointsize, alpha=alpha, edgecolor="black")
    plt.axis('tight')
    plt.xticks([])
    plt.yticks([])
    # add mapping from mean positions to modes
    # (GTM nodes where the data points have max probability)
    for i in range(means.shape[0]):
        plt.plot([means[i, 0], modes[i, 0]],
                 [means[i, 1], modes[i, 1]], color='grey', linewidth=0.5)

    # plot4: GTM landscape visualization without means/modes mappings
    fig.add_subplot(224)
    if discrete and labels is not None:
        plotClassMapNoPoints(optimizedModel, label_numeric,
                             do_interpolate=do_interpolate, cname=cname)
    elif labels is not None:
        plotLandscapeNoPoints(optimizedModel, label_numeric,
                              do_interpolate=do_interpolate,cname=cname)
    fig.set_size_inches(16, 13)
    fig.savefig(output+".pdf", format='pdf', dpi=500)
    plt.close(fig)
    print("\nWrote pdf to disk: %s\n" % (output+".pdf"))


def plot_html(coordinates, labels=None, ids=None, title="plot",
              discrete=False, output="output", pointsize=1.0, alpha=0.3,
              cname="Spectral_r"):
    if labels is None:
        colvec = "black"
    elif discrete:
        uniqClasses, label_numeric = np.unique(labels, return_inverse=True)
        colvec = np.squeeze(label_numeric)
    elif not discrete:
        label_numeric = labels
        colvec = np.squeeze(label_numeric)
    fig, ax = plt.subplots(subplot_kw=dict(facecolor='#EEEEEE'))
    ax.grid(color='white', linestyle='solid')
    ax.set_title(title, size=30)
    pointsidx = ['point {0}'.format(i + 1)
                 for i in range(coordinates.shape[0])]
    scatter = ax.scatter(coordinates[:, 0].tolist(),
                         coordinates[:, 1].tolist(),
                         c=colvec, s=20*pointsize, alpha=alpha,
                         cmap=plt.get_cmap(cname),
                         edgecolor='black')
    if ids is not None and labels is not None:
        tooltipstr = ["%s: label=%s" % t for t in zip(ids, labels)]
    elif ids is not None:
        tooltipstr = ["%s: id=%s" % t for t in zip(pointsidx, ids)]
    elif labels is not None:
        tooltipstr = ["%s: label=%s" % t for t in zip(pointsidx, labels)]
    else:
        tooltipstr = list(pointsidx)
    tooltip = mpld3.plugins.PointLabelTooltip(scatter, labels=tooltipstr)
    mpld3.plugins.connect(fig, tooltip)
    mpld3.save_html(fig, output+".html")
    plt.close(fig)
    print("\nWrote html plot to disk: %s\n" % (output+".html"))


def plot_html_GTM(optimizedModel, labels=None, ids=None, plot_arrows=True,
                  title="GTM", discrete=False, output="output",
                  pointsize=1.0, alpha=0.3, do_interpolate=True,
                  cname="Spectral_r"):
    if labels is None:
        colvec = "black"
    elif discrete:
        uniqClasses, label_numeric = np.unique(labels, return_inverse=True)
        colvec = np.squeeze(label_numeric)
    elif not discrete:
        label_numeric = labels
        colvec = np.squeeze(label_numeric)
    fig, ax = plt.subplots(subplot_kw=dict(facecolor='#EEEEEE'))
    if discrete and labels is not None:
        plotClassMapNoPoints(optimizedModel, label_numeric,
                             do_interpolate=do_interpolate, cname=cname)
    elif labels is not None:
        plotLandscapeNoPoints(optimizedModel, label_numeric,
                              do_interpolate=do_interpolate, cname=cname)
    means = optimizedModel.matMeans
    modes = optimizedModel.matModes
    ax.grid(color='white', linestyle='solid')
    ax.set_title(title, size=30)
    pointsidx = ['point {0}'.format(i + 1) for i in range(means.shape[0])]
    scatter = ax.scatter(means[:, 0], means[:, 1], c=colvec,
                         s=20*pointsize,
                         alpha=alpha, cmap=plt.get_cmap(cname),
                         edgecolor='black')
    if plot_arrows:
        for i in range(means.shape[0]):
            plt.plot([means[i, 0], modes[i, 0]],
                     [means[i, 1], modes[i, 1]], color='grey', linewidth=0.5)
    if ids is not None and labels is not None:
        tooltipstr = ["%s: label=%s" % t for t in zip(ids, labels)]
    elif ids is not None:
        tooltipstr = ["%s: id=%s" % t for t in zip(pointsidx, ids)]
    elif labels is not None:
        tooltipstr = ["%s: label=%s" % t for t in zip(pointsidx, labels)]
    else:
        tooltipstr = list(pointsidx)
    tooltip = mpld3.plugins.PointLabelTooltip(scatter, labels=tooltipstr)
    mpld3.plugins.connect(fig, tooltip)
    mpld3.save_html(fig, output+".html")
    plt.close(fig)
    print("\nWrote html plot to disk: %s\n" % (output+".html"))


def plot_html_GTM_projection(optimizedModel, projections, labels=None,
                             ids=None, plot_arrows=True,
                             title="GTM_projection",
                             discrete=False, output="output",
                             pointsize=1, alpha=0.3,
                             do_interpolate=True, cname="Spectral_r"):
    if discrete:
        uniqClasses, label_numeric = np.unique(labels, return_inverse=True)
    elif not discrete:
        label_numeric = labels
    fig, ax = plt.subplots(subplot_kw=dict(facecolor='#EEEEEE'))
    if discrete and labels is not None:
        plotClassMap(optimizedModel, np.squeeze(label_numeric),
                     do_interpolate=do_interpolate, cname=cname,
                     pointsize=pointsize, alpha=alpha)
    elif labels is not None:
        plotLandscape(optimizedModel, np.squeeze(label_numeric),
                      do_interpolate=do_interpolate, cname=cname,
                      pointsize=pointsize, alpha=alpha)
    means = projections.matMeans
    modes = projections.matModes
    ax.grid(color='white', linestyle='solid')
    ax.set_title(title, size=30)
    pointsidx = ['point {0}'.format(i + 1) for i in range(means.shape[0])]
    scatter = ax.scatter(means[:, 0], means[:, 1], c="black",
                         s=20*pointsize, alpha=alpha, edgecolor='black')
    if plot_arrows:
        for i in range(means.shape[0]):
            plt.plot([means[i, 0], modes[i, 0]],
                     [means[i, 1], modes[i, 1]], color='grey', linewidth=0.5)
    if ids is not None:
        tooltip = mpld3.plugins.PointLabelTooltip(scatter, labels=list(ids))
    else:
        tooltip = mpld3.plugins.PointLabelTooltip(
            scatter, labels=list(pointsidx))
    mpld3.plugins.connect(fig, tooltip)
    mpld3.save_html(fig, output+".html")
    plt.close(fig)
    print("\nWrote html plot to disk: %s\n" % (output+".html"))


def plotLandscape(optimizedModel, labels, do_interpolate=True,
                  cname="Spectral_r", pointsize=1.0, alpha=0.3):
    k = math.sqrt(optimizedModel.matX.shape[0])
    n = 100
    x = optimizedModel.matX[:, 0]
    y = optimizedModel.matX[:, 1]
    z = ugtm_landscape.landscape(optimizedModel, labels)
    if do_interpolate:
        ti = np.linspace(-1.1, 1.1, n)
        XI, YI = np.meshgrid(ti, ti)
        f = interpolate.NearestNDInterpolator(optimizedModel.matX, z)
        ZI = f(XI, YI)
        plt.pcolor(XI, YI, ZI, cmap=plt.get_cmap(cname))
    else:
        plt.scatter(x, y, 50*(10/k), z, cmap=plt.get_cmap(cname), marker="s")
    plt.scatter(optimizedModel.matMeans[:, 0], optimizedModel.matMeans[:, 1],
                s=pointsize*20, c=np.squeeze(labels), cmap=plt.get_cmap(cname),
                edgecolor='black', marker="o")
    plt.title('Landscape')
    plt.xlim(-1.1, 1.1)
    plt.ylim(-1.1, 1.1)
    plt.colorbar()
    plt.axis('tight')
    plt.xticks([]), plt.yticks([])


def plotLandscapeNoPoints(optimizedModel, labels, do_interpolate=True,
                          cname="Spectral_r"):
    k = math.sqrt(optimizedModel.matX.shape[0])
    n = 100
    x = optimizedModel.matX[:, 0]
    y = optimizedModel.matX[:, 1]
    z = ugtm_landscape.landscape(optimizedModel, labels)
    if do_interpolate:
        ti = np.linspace(-1.1, 1.1, n)
        XI, YI = np.meshgrid(ti, ti)
        f = interpolate.NearestNDInterpolator(optimizedModel.matX, z)
        ZI = f(XI, YI)
        plt.pcolor(XI, YI, ZI, cmap=plt.get_cmap(cname))
    else:
        plt.scatter(x, y, 50*(10/k), z, cmap=plt.get_cmap(cname), marker="s")
    plt.title('Landscape')
    plt.xlim(-1.1, 1.1)
    plt.ylim(-1.1, 1.1)
    plt.colorbar()
    plt.axis('tight')
    plt.xticks([]), plt.yticks([])


def plotClassMap(optimizedModel, labels, prior="equiprobable",
                 do_interpolate=True, cname="Spectral_r", pointsize=1.0,
                 alpha=0.3):
    k = math.sqrt(optimizedModel.matX.shape[0])
    n = 100
    x = optimizedModel.matX[:, 0]
    y = optimizedModel.matX[:, 1]
    uniqClasses, label_numeric = np.unique(labels, return_inverse=True)
    z = ugtm_landscape.classMap(optimizedModel,
                                label_numeric, prior).activityModel
    if do_interpolate:
        ti = np.linspace(-1.1, 1.1, n)
        XI, YI = np.meshgrid(ti, ti)
        f = interpolate.NearestNDInterpolator(optimizedModel.matX, z)
        ZI = f(XI, YI)
        plt.pcolor(XI, YI, ZI, cmap=plt.get_cmap(cname))
    else:
        plt.scatter(x, y, 175*(10/k), z, cmap=plt.get_cmap(cname),
                    marker="s", alpha=0.3)
    plt.scatter(optimizedModel.matMeans[:, 0], optimizedModel.matMeans[:, 1],
                s=20*pointsize, c=np.squeeze(label_numeric),
                cmap=plt.get_cmap(cname),
                edgecolor='black', marker="o")
    plt.title('Class Map')
    plt.xlim(-1.1, 1.1)
    plt.ylim(-1.1, 1.1)
    plt.axis('tight')
    plt.xticks([]), plt.yticks([])


def plotClassMapNoPoints(optimizedModel, labels, prior="equiprobable",
                         do_interpolate=True, cname="Spectral_r"):
    k = math.sqrt(optimizedModel.matX.shape[0])
    n = 100
    x = optimizedModel.matX[:, 0]
    y = optimizedModel.matX[:, 1]
    uniqClasses, label_numeric = np.unique(labels, return_inverse=True)
    z = ugtm_landscape.classMap(optimizedModel,
                                label_numeric, prior).activityModel
    if do_interpolate:
        ti = np.linspace(-1.1, 1.1, n)
        XI, YI = np.meshgrid(ti, ti)
        f = interpolate.NearestNDInterpolator(optimizedModel.matX, z)
        ZI = f(XI, YI)
        plt.pcolor(XI, YI, ZI, cmap=plt.get_cmap(cname))
    else:
        plt.scatter(x, y, 175*(10/k), z, cmap=plt.get_cmap(cname),
                    marker="s", alpha=0.3)
    plt.title('Class Map')
    plt.xlim(-1.1, 1.1)
    plt.ylim(-1.1, 1.1)
    plt.axis('tight')
    plt.xticks([]), plt.yticks([])



from __future__ import print_function
import numpy as np
from .ugtm_plot import plot_pdf
from .ugtm_plot import plotMultiPanelGTM
from .ugtm_plot import plot_html_GTM
from .ugtm_plot import plot_html_GTM_projection


class ReturnU(object):
    def __init__(self, matU, betaInv):
        self.matU = matU
        self.betaInv = betaInv


class InitialGTM(object):
    def __init__(self, matX, matM, n_nodes, n_rbf_centers, rbfWidth,
                 matPhiMPlusOne, matW, matY, betaInv, n_dimensions):
        self.matX = matX
        self.matM = matM
        self.n_rbf_centers = n_rbf_centers
        self.n_nodes = n_nodes
        self.rbfWidth = rbfWidth
        self.matPhiMPlusOne = matPhiMPlusOne
        self.matW = matW
        self.matY = matY
        self.betaInv = betaInv
        self.n_dimensions = n_dimensions


class OptimizedGTM(object):
    def __init__(self, matW, matY, matP, matR, betaInv, matMeans,
                 matModes, matX, n_dimensions, converged):
        self.matW = matW
        self.matY = matY
        self.matP = matP
        self.matR = matR
        self.betaInv = betaInv
        self.matMeans = matMeans
        self.matModes = matModes
        self.matX = matX
        self.n_dimensions = n_dimensions
        self.converged = converged

    def write(self, output="output"):
        np.savetxt(fname=output+"_responsibilities.csv",
                   X=self.matR, delimiter=",")
        np.savetxt(fname=output+"_coordinates.csv",
                   X=self.matMeans, delimiter=",")
        np.savetxt(fname=output+"_modes.csv", X=self.matModes, delimiter=",")
        print("")
        print("Wrote to disk:")
        print("")
        print("%s: responsibilities, which represent "
              "each individual's encoding "
              "on the map (dimensions=n_individuals*n_nodes_on_the_map)"
              % (output+"_responsibilities.csv"))
        print("")
        print("%s: coordinates to plot, which represent each individual's "
              "mean position on the map (dimensions = "
              "n_individuals*n_latent_dimensions)"
              % (output+"_coordinates.csv"))
        print("")
        print("%s: modes positions for each individual on the map "
              "(node with max probability for the individual; "
              "dimensions = n_individuals*n_latent_dimensions)"
              % (output+"_modes.csv"))
        print("")
        print("")

    def write_all(self, output="output"):
        outparams = "n_dimensions:"+str(self.n_dimensions) + \
                    "\n"+"inverse_variance:"+str(self.betaInv)
        np.savetxt(fname=output+"_responsibilities.csv",
                   X=self.matR, delimiter=",")
        np.savetxt(fname=output+"_coordinates.csv",
                   X=self.matMeans, delimiter=",")
        np.savetxt(fname=output+"_modes.csv", X=self.matModes, delimiter=",")
        np.savetxt(fname=output+"_manifold.csv", X=self.matY, delimiter=",")
        np.savetxt(fname=output+"_parametersMatrix.csv",
                   X=self.matW, delimiter=",")
        np.savetxt(fname=output+"_dimensionsAndVariance.csv", X=outparams)
        print("")
        print("Wrote to disk:")
        print("")
        print("%s: responsibilities, which represent "
              "each individual's encoding on the map "
              "(dimensions=n_individuals*n_nodes_on_the_map)"
              % (output+"_responsibilities.csv"))
        print("")
        print("%s: coordinates to plot, which represent each individual's "
              "mean position on the map "
              "(dimensions = n_individuals*n_latent_dimensions)"
              % (output+"_coordinates.csv"))
        print("")
        print("%s: modes positions for each individual on the map "
              "(node with max probability for the individual; "
              "dimensions = n_individuals*n_latent_dimensions)"
              % (output+"_modes.csv"))
        print("")
        print("%s: manifold coordinates in the initial data space "
              "(dimensions: n_data_dimensions*n_points_on_manifold"
              % (output+"_manifold.csv"))
        print("")
        print("%s: parameters matrix"
              % (output+"_parametersMatrix.csv"))
        print("")
        print("%s: initial space and inverse variance"
              % (output+"_dimensionsAndVariance.csv"))
        print("")
        print("")

    def plot(self, labels=None, title="", output="output",
             discrete=False, pointsize=1, alpha=0.3, cname="Spectral_r"):
        plot_pdf(coordinates=self.matMeans, labels=labels, title=title,
                 output=output, discrete=discrete,
                 pointsize=pointsize, alpha=alpha, cname=cname)

    def plot_modes(self, labels=None, title="", output="output",
                   discrete=False, pointsize=1, alpha=0.3, cname="Spectral_r"):
        plot_pdf(coordinates=self.matModes, labels=labels, title=title,
                 output=output, discrete=discrete,
                 pointsize=pointsize, alpha=alpha, cname=cname)

    def plot_html(self, labels=None, ids=None, plot_arrows=True,
                  title="GTM", discrete=False, output="output",
                  pointsize=1.0, alpha=0.3, do_interpolate=True,
                  cname="Spectral_r"):
        plot_html_GTM(optimizedModel=self, labels=labels, ids=ids,
                      plot_arrows=plot_arrows, title=title, discrete=discrete,
                      output=output, pointsize=pointsize, alpha=alpha,
                      do_interpolate=do_interpolate, cname=cname)

    def plot_multipanel(self, labels, output="output", discrete=False,
                        pointsize=1.0, alpha=0.3, do_interpolate=True,
                        cname="Spectral_r"):
        plotMultiPanelGTM(optimizedModel=self, labels=labels, output=output,
                          discrete=discrete, pointsize=pointsize, alpha=alpha,
                          do_interpolate=do_interpolate, cname=cname)

    def plot_html_projection(self, projections, labels=None,
                             ids=None, plot_arrows=True,
                             title="GTM_projection", discrete=False,
                             output="output", pointsize=1.0,
                             alpha=0.3, do_interpolate=True,
                             cname="Spectral_r"):
        plot_html_GTM_projection(optimizedModel=self, projections=projections,
                                 labels=labels, ids=ids,
                                 plot_arrows=plot_arrows,
                                 title=title, discrete=discrete,
                                 output=output,
                                 pointsize=pointsize, alpha=alpha,
                                 do_interpolate=do_interpolate, cname=cname)



from __future__ import print_function
import numpy as np
from sklearn.decomposition import PCA
from . import ugtm_preprocess
from . import ugtm_classes
from . import ugtm_core


def initializeKernel(data, k, m, s, maxdim, random_state=1234):
    n_individuals = data.shape[0]
    n_nodes = k*k
    n_rbf_centers = m*m
    x = np.linspace(-1, 1, k)
    matX = np.transpose(np.meshgrid(x, x)).reshape(k*k, 2)
    x = np.linspace(-1, 1, m)
    matM = np.transpose(np.meshgrid(x, x)).reshape(m*m, 2)
    rbfWidth = ugtm_core.computeWidth(matM, n_rbf_centers, s)
    matPhiMPlusOne = ugtm_core.createPhiMatrix(
        matX, matM, n_nodes, n_rbf_centers, rbfWidth)
    pca = PCA(random_state=random_state)
    pca.fit(data)
    matW = (pca.components_.T * np.sqrt(pca.explained_variance_)
            )[:, 0:n_rbf_centers+1]
    n_dimensions = np.searchsorted(
        pca.explained_variance_ratio_.cumsum(), 0.995)+1
    if n_dimensions > maxdim:
        n_dimensions = maxdim
    # betaInv=pca.explained_variance_[2]
    matD = ugtm_core.KERNELcreateDistanceMatrix(data, matW, matPhiMPlusOne)
    betaInv = ugtm_core.initBetaInvRandom(matD, n_nodes, n_individuals,
                                          n_dimensions)
    matY = ugtm_core.createYMatrixInit(data, matW, matPhiMPlusOne)
    return ugtm_classes.InitialGTM(matX, matM, n_nodes, n_rbf_centers,
                                   rbfWidth,
                                   matPhiMPlusOne, matW,
                                   matY, betaInv, n_dimensions)


def optimizeKernel(data, initialModel, l, niter, verbose=True):
    matD = ugtm_core.KERNELcreateDistanceMatrix(
        data, initialModel.matW, initialModel.matPhiMPlusOne)
    matY = initialModel.matY
    betaInv = initialModel.betaInv
    i = 1
    diff = 1000
    converged = 0
    while (i < (niter+1)) and (converged < 4):
        # expectation
        matP = ugtm_core.createPMatrix(matD, betaInv,
                                       initialModel.n_dimensions)
        matR = ugtm_core.createRMatrix(matP)
        # maximization
        matG = ugtm_core.createGMatrix(matR)
        matW = ugtm_core.optimLMatrix(
            matR, initialModel.matPhiMPlusOne, matG, betaInv, l)
        matY = ugtm_core.createYMatrix(matW, initialModel.matPhiMPlusOne)
        matD = ugtm_core.KERNELcreateDistanceMatrix(
            data, matW, initialModel.matPhiMPlusOne)
        betaInv = ugtm_core.optimBetaInv(matR, matD, initialModel.n_dimensions)
        # objective function
        if i == 1:
            loglike = ugtm_core.computelogLikelihood(
                matP, betaInv, initialModel.n_dimensions)
        else:
            loglikebefore = loglike
            loglike = ugtm_core.computelogLikelihood(
                matP, betaInv, initialModel.n_dimensions)
            diff = abs(loglikebefore-loglike)
        if verbose is True:
            print("Iter ", i, " Err: ", loglike)
        if diff <= 0.0001:
            converged += 1
        else:
            converged = 0
        i += 1
    if verbose is True:
        if converged >= 3:
            print("Converged: ", loglike)
    if converged >= 3:
        has_converged = True
    else:
        has_converged = False
    matY = ugtm_core.createYMatrix(matW, initialModel.matPhiMPlusOne)
    matMeans = ugtm_core.meanPoint(matR, initialModel.matX)
    matModes = ugtm_core.modePoint(matR, initialModel.matX)
    return ugtm_classes.OptimizedGTM(matW, matY, matP.T, matR.T,
                                     betaInv, matMeans,
                                     matModes,
                                     initialModel.matX,
                                     initialModel.n_dimensions,
                                     has_converged)


def runkGTM(data, doPCA=False, doKernel=False, kernel="linear",
            n_components=-1,
            maxdim=100, missing=True, missing_strategy="median",
            random_state=1234, k=16, m=4, s=0.3, l=0.1, niter=200,
            verbose=False):
    data = ugtm_preprocess.pcaPreprocess(data=data, doPCA=doPCA,
                                         n_components=n_components,
                                         missing=missing,
                                         missing_strategy=missing_strategy,
                                         random_state=random_state)
    if doKernel or data.shape[0] != data.shape[1]:
        data = ugtm_preprocess.chooseKernel(data, kernel)
    initialModel = initializeKernel(data, k, m, s, random_state)
    optimizedModel = optimizeKernel(
        data, initialModel, l, niter, verbose=verbose)
    return optimizedModel



from __future__ import print_function
import numpy as np
from sklearn.metrics import pairwise_distances
from sklearn.metrics import pairwise
from scipy.spatial import distance
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import KernelCenterer
from sklearn.decomposition import PCA
from sklearn.preprocessing import Imputer


class ProcessedTrainTest(object):
    def __init__(self, train, test):
        self.train = train
        self.test = test


def pcaPreprocess(data, doPCA=False, n_components=-1, missing=False,
                  missing_strategy='most_frequent', random_state=1234):
    if missing:
        imp = Imputer(strategy=missing_strategy, axis=0)
        data = imp.fit_transform(data)
    scaler = StandardScaler()
    data = scaler.fit_transform(data)
    if n_components == -1:
        n_components = 0.80
    if doPCA:
        pca = PCA(random_state=random_state, n_components=n_components)
        data = pca.fit_transform(data)
        n_components = pca.n_components_
        print("Used %s components explaining %s%% of the variance\n" %
              (n_components,
               pca.explained_variance_ratio_.cumsum()[n_components-1]*100))
    return(data)


def processTrainTest(train, test, doPCA, n_components, missing=False,
                     missing_strategy='most_frequent', random_state=1234):
    if missing:
        imp = Imputer(strategy=missing_strategy, axis=0)
        train = imp.fit_transform(train)
        test = imp.transform(test)
    scaler = StandardScaler()
    train = scaler.fit_transform(train)
    test = scaler.transform(test)
    if(n_components == -1):
        n_components = 0.80
    if doPCA:
        pca = PCA(random_state=random_state, n_components=n_components)
        train = pca.fit_transform(train)
        test = pca.transform(test)
    return(ProcessedTrainTest(train, test))


def chooseKernel(data, kerneltype='euclidean'):
    if kerneltype == 'euclidean':
        K = np.divide(1, (1+pairwise_distances(data, metric="euclidean")))
    elif kerneltype == 'cosine':
        K = (pairwise.cosine_kernel(data))
    elif kerneltype == 'laplacian':
        K = (pairwise.laplacian_kernel(data))
    elif kerneltype == 'linear':
        K = (pairwise.linear_kernel(data))
    elif kerneltype == 'polynomial_kernel':
        K = (pairwise.polynomial_kernel(data))
    elif kerneltype == 'jaccard':
        K = 1-distance.cdist(data, data, metric='jaccard')
    scaler = KernelCenterer().fit(K)
    return(scaler.transform(K))



from __future__ import print_function
import numpy as np
from sklearn.decomposition import PCA
from . import ugtm_preprocess
from . import ugtm_classes
from . import ugtm_core


def initialize(data, k, m, s, random_state=1234):
    n_dimensions = data.shape[1]
    n_nodes = k*k
    n_rbf_centers = m*m
    x = np.linspace(-1, 1, k)
    matX = np.transpose(np.meshgrid(x, x)).reshape(k*k, 2)
    x = np.linspace(-1, 1, m)
    matM = np.transpose(np.meshgrid(x, x)).reshape(m*m, 2)
    rbfWidth = ugtm_core.computeWidth(matM, n_rbf_centers, s)
    matPhiMPlusOne = ugtm_core.createPhiMatrix(
        matX, matM, n_nodes, n_rbf_centers, rbfWidth)
    pca = PCA(n_components=3, random_state=random_state)
    pca.fit(data)
    matU = (pca.components_.T * np.sqrt(pca.explained_variance_))[:, 0:2]
    betaInv = pca.explained_variance_[2]
    Uobj = ugtm_classes.ReturnU(matU, betaInv)
    matW = ugtm_core.createWMatrix(matX, matPhiMPlusOne, Uobj.matU,
                                   n_dimensions, n_rbf_centers)
    matY = ugtm_core.createYMatrixInit(data, matW, matPhiMPlusOne)
    betaInv = Uobj.betaInv
    betaInv = ugtm_core.evalBetaInv(matY, Uobj.betaInv,
                                    random_state=random_state)
    return ugtm_classes.InitialGTM(matX, matM, n_nodes,
                                   n_rbf_centers, rbfWidth,
                                   matPhiMPlusOne,
                                   matW, matY, betaInv, n_dimensions)


def optimize(data, initialModel, l, niter, verbose=True):
    matD = ugtm_core.createDistanceMatrix(initialModel.matY, data)
    matY = initialModel.matY
    betaInv = initialModel.betaInv
    i = 1
    diff = 1000
    converged = 0
    while i < (niter+1) and (converged < 4):
        # expectation
        matP = ugtm_core.createPMatrix(matD, betaInv,
                                       initialModel.n_dimensions)
        matR = ugtm_core.createRMatrix(matP)
        # maximization
        matG = ugtm_core.createGMatrix(matR)
        matW = ugtm_core.optimWMatrix(
            matR, initialModel.matPhiMPlusOne, matG, data, betaInv, l)
        matY = ugtm_core.createYMatrix(matW, initialModel.matPhiMPlusOne)
        matD = ugtm_core.createDistanceMatrix(matY, data)
        betaInv = ugtm_core.optimBetaInv(
            matR, matD, initialModel.n_dimensions)
        # objective function
        if i == 1:
            loglike = ugtm_core.computelogLikelihood(
                matP, betaInv, initialModel.n_dimensions)
        else:
            loglikebefore = loglike
            loglike = ugtm_core.computelogLikelihood(
                matP, betaInv, initialModel.n_dimensions)
            diff = abs(loglikebefore-loglike)
        if diff <= 0.0001:
            converged += 1
        else:
            converged = 0
        if verbose:
            print("Iter ", i, " Err: ", loglike)
        i += 1
    # final iteration to make sure matR fits matD
    if verbose == 1:
        if converged >= 3:
            print("Converged: ", loglike)
    if converged >= 3:
        has_converged = True
    else:
        has_converged = False
    matP = ugtm_core.createPMatrix(matD, betaInv, initialModel.n_dimensions)
    matR = ugtm_core.createRMatrix(matP)
    matMeans = ugtm_core.meanPoint(matR, initialModel.matX)
    matModes = ugtm_core.modePoint(matR, initialModel.matX)
    return ugtm_classes.OptimizedGTM(matW, matY, matP.T, matR.T,
                                     betaInv, matMeans,
                                     matModes,
                                     initialModel.matX,
                                     initialModel.n_dimensions,
                                     has_converged)


def projection(optimizedModel, new_data):
    matD = ugtm_core.createDistanceMatrix(optimizedModel.matY, new_data)
    matP = ugtm_core.createPMatrix(matD, optimizedModel.betaInv,
                                   optimizedModel.n_dimensions)
    matR = ugtm_core.createRMatrix(matP)
    # loglike = ugtm_core.computelogLikelihood(
    # matP, optimizedModel.betaInv, optimizedModel.n_dimensions)
    matMeans = ugtm_core.meanPoint(matR, optimizedModel.matX)
    matModes = ugtm_core.modePoint(matR, optimizedModel.matX)
    return ugtm_classes.OptimizedGTM(optimizedModel.matW, optimizedModel.matY,
                                     matP.T, matR.T,
                                     optimizedModel.betaInv,
                                     matMeans, matModes,
                                     optimizedModel.matX,
                                     optimizedModel.n_dimensions,
                                     optimizedModel.converged)


def transform(optimizedModel, train, test, doPCA=False, n_components=-1,
              missing=True,
              missing_strategy="median", random_state=1234, process=True):
    if process is True:
        processed = ugtm_preprocess.processTrainTest(train, test, doPCA=doPCA,
                                                     n_components=n_components,
                                                     missing=missing,
                                                     missing_strategy=missing_strategy,
                                                     random_state=random_state)
    if process is True:
        return projection(optimizedModel, processed.test)
    else:
        return projection(optimizedModel, test)


def runGTM(data, doPCA=False, n_components=-1,
           missing=True, missing_strategy="median",
           random_state=1234, k=16, m=4, s=0.3, l=0.1,
           niter=200, verbose=False):
    data = ugtm_preprocess.pcaPreprocess(data=data, doPCA=doPCA,
                                         n_components=n_components,
                                         missing=missing,
                                         missing_strategy=missing_strategy,
                                         random_state=random_state)
    initialModel = initialize(data, k, m, s, random_state)
    optimizedModel = optimize(data, initialModel, l, niter, verbose=verbose)
    return optimizedModel



from __future__ import print_function
import numpy as np


class ClassMap(object):
    def __init__(self, nodeClassP, nodeClassT, activityModel, uniqClasses):
        self.nodeClassP = nodeClassP
        self.nodeClassT = nodeClassT
        self.activityModel = activityModel
        self.uniqClasses = uniqClasses


def landscape(optimizedModel, activity):
    epsilon = 10e-8
    sums = np.sum(optimizedModel.matR+epsilon, axis=0)
    landscape = np.dot(activity.flatten(),
                       optimizedModel.matR+epsilon) / sums[None, :]
    return np.asarray(landscape)[0, :]


def classMap(optimizedModel, activity, prior="equiprobable"):
    uniqClasses, classVector = np.unique(activity, return_inverse=True)
    nClasses = uniqClasses.shape[0]
    n_nodes = optimizedModel.matR.shape[1]
    # posterior distribution
    nodeClassP = np.zeros([n_nodes, nClasses])
    # likelihood
    nodeClassT = np.zeros([n_nodes, nClasses])
    sumClass = np.zeros([nClasses])
    summe = np.zeros([n_nodes])
    for i in range(nClasses):
        sumClass[i] = (classVector == i).sum()
    if prior == "estimated":
        priors = sumClass/sumClass.sum()
    elif prior == "equiprobable":
        priors = np.zeros([nClasses])+(1.0/nClasses)

    for i in range(nClasses):
        for k in range(n_nodes):
            nodeClassT[k, i] = optimizedModel.matR[classVector ==
                                                   i, k].sum()/sumClass[i]

    for i in range(nClasses):
        for k in range(n_nodes):
            nodeClassP[k, i] = nodeClassT[k, i]*priors[i]
            summe[k] += nodeClassP[k, i]

    for i in range(nClasses):
        for k in range(n_nodes):
            if summe[k] != 0.0:
                nodeClassP[k, i] = nodeClassP[k, i]/summe[k]

    for k in range(n_nodes):
        if summe[k] == 0.0:
            for i in range(nClasses):
                nodeClassP[k, i] = 1/nClasses

    nodeClass = np.argmax(nodeClassP, axis=1)
    return(ClassMap(nodeClassP, nodeClassT, nodeClass, uniqClasses))



from __future__ import print_function
import numpy as np
from scipy.spatial import distance
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.preprocessing import scale
# create manifold


def createYMatrixInit(data, matW, matPhiMPlusOne):
    shap1 = matW.shape[0]
    shap2 = matPhiMPlusOne.shape[0]
    TheMeans = data.mean(0)
    DMmeanMatrix = np.zeros([shap1, shap2])
    for i in range(shap1):
        for j in range(shap2):
            DMmeanMatrix[i, j] = TheMeans[i]
    MatY = np.dot(matW, np.transpose(matPhiMPlusOne))
    MatY = MatY + DMmeanMatrix
    return(MatY)


def createPhiMatrix(matX, matM, numX, numM, sigma):
    Result = np.zeros([numX, numM + 1])
    for i in range(numX):
        for j in range(numM):
            Coo1 = (matX[i][0] - matM[j][0]) * (matX[i][0] - matM[j][0])
            Coo2 = (matX[i][1] - matM[j][1]) * (matX[i][1] - matM[j][1])
            Distance = Coo1 + Coo2
            Result[i, j] = np.exp(-(Distance) / (2 * sigma))
    for i in range(numX):
        Result[i][numM] = 1
    return(Result)


def computeWidth(matM, numM, sigma):
    Result = 0.0
    Distances = np.zeros([numM, numM])
    mins = np.zeros([numM, 1])
    maxs = np.zeros([numM, 1])
    for i in range(numM):
        for j in range(numM):
            Coo1 = (matM[i][0] - matM[j][0]) * (matM[i][0] - matM[j][0])
            Coo2 = (matM[i][1] - matM[j][1]) * (matM[i][1] - matM[j][1])
            Distances[i, j] = Coo1 + Coo2
    for i in range(numM):
        mins[i] = np.min(Distances[i][np.nonzero(Distances[i])])
    for i in range(numM):
        maxs[i] = np.max(Distances[i][np.nonzero(Distances[i])])
    if (sigma > 0.0):
        Result = sigma * np.mean(mins)
    else:
        Result = np.max(maxs)
    return(Result)


def createWMatrix(matX, matPhiMPlusOne, matU, n_dimensions, n_rbf_centers):
    NormX = scale(matX, axis=0, with_mean=True, with_std=True)
    myProd = np.dot(matU, np.transpose(NormX))
    tinv = np.linalg.solve(matPhiMPlusOne.T.dot(
        matPhiMPlusOne), matPhiMPlusOne.T)
    Result = np.dot(myProd, np.transpose(tinv))
    return(Result)


def createYMatrix(matW, matPhiMPlusOne):
    Result = np.dot(matW, np.transpose(matPhiMPlusOne))
    return(Result)


def createDistanceMatrix(matY, data):
    Result = distance.cdist(matY.T, data, metric='sqeuclidean')
    return(Result)


def KERNELcreateDistanceMatrix(data, matL, matPhiMPlusOne):
    n_nodes = matPhiMPlusOne.shape[0]
    n_individuals = data.shape[0]
    Result = np.zeros([n_nodes, n_individuals])
    thefloat = 0.0
    for i in range(n_nodes):
        LPhim = np.dot(matL, matPhiMPlusOne[i])
        thefloat = np.dot(np.dot(LPhim, data), LPhim)
        for j in range(n_individuals):
            Result[i, j] = data[j, j] + thefloat - 2*(np.dot(data[j], LPhim))
    return(Result)


def exp_normalize(x):
    y = np.array([], dtype=np.longdouble)
    y = x - np.expand_dims(np.max(x, axis=0), 0)
    y = np.exp(y)
    return(y)


def createPMatrix(matD, betaInv, n_dimensions):
    matP = np.array([], dtype=np.longdouble)
    beta = 1/betaInv
    # constante = np.power(((beta)/(2*np.pi)),dim/2)
    # constante = 1
    matP = exp_normalize(-(beta/2)*matD)
    return(matP)


def createRMatrix(matP):
    matR = np.array([], dtype=np.longdouble)
    sums = np.sum(matP, axis=0)
    matR = (matP) / (sums[None, :])
    return(matR)


def createGMatrix(matR):
    sums = np.sum(matR, axis=1)
    matG = np.diag(sums)
    return(matG)


def optimWMatrix(matR, matPhiMPlusOne, matG, data, betaInv, l):
    n_rbf_centersP = matPhiMPlusOne.shape[1]
    LBmat = np.zeros([n_rbf_centersP, n_rbf_centersP])
    PhiGPhi = np.dot(
        np.dot(np.transpose(matPhiMPlusOne), matG), matPhiMPlusOne)
    for i in range(n_rbf_centersP):
        LBmat[i][i] = l * betaInv
    PhiGPhiLB = PhiGPhi + LBmat
    Ginv = np.linalg.inv(PhiGPhiLB)
    matW = np.transpose(
        np.dot(np.dot(np.dot(Ginv, np.transpose(matPhiMPlusOne)), matR), data))
    return(matW)


def optimLMatrix(matR, matPhiMPlusOne, matG, betaInv, l):
    n_rbf_centersP = matPhiMPlusOne.shape[1]
    LBmat = np.zeros([n_rbf_centersP, n_rbf_centersP])
    PhiGPhi = np.dot(
        np.dot(np.transpose(matPhiMPlusOne), matG), matPhiMPlusOne)
    for i in range(n_rbf_centersP):
        LBmat[i][i] = l * betaInv
    PhiGPhiLB = PhiGPhi + LBmat
    Ginv = np.linalg.inv(PhiGPhiLB)
    matW = np.transpose(
        np.dot(np.dot(Ginv, np.transpose(matPhiMPlusOne)), matR))
    return(matW)


def optimBetaInv(matR, matD, n_dimensions):
    n_individuals = matR.shape[1]
    betaInv = np.sum(np.multiply(matR, matD))/(n_individuals*n_dimensions)
    return(betaInv)


def meanPoint(matR, matX):
    matMeans = np.dot(np.transpose(matR), matX)
    return(matMeans)


def modePoint(matR, matX):
    matModes = matX[np.argmax(matR, axis=0), :]
    return(matModes)


def computelogLikelihood(matP, betaInv, n_dimensions):
    LogLikelihood = np.longdouble()
    n_nodes = matP.shape[0]
    n_individuals = matP.shape[1]
    LogLikelihood = 0.0
    prior = 1.0/n_nodes
    placeholder = 50
    constante = np.longdouble(np.power(((1/betaInv)/(2*np.pi)),
                                       np.minimum(
                                        n_dimensions/2, placeholder)))
    LogLikelihood = np.sum(np.log(
                            np.maximum(np.sum(constante*matP, axis=0)*prior,
                                       np.finfo(np.longdouble).tiny)))
    LogLikelihood /= n_individuals
    return(-LogLikelihood)


def evalBetaInv(matY, betaInv, random_state=1234):
    r = np.random.RandomState(random_state)
    Distances = euclidean_distances(np.transpose(matY))
    myMin = np.mean(Distances)/2
    myMin *= myMin
    if((myMin < betaInv) or (betaInv == 0)):
        betaInv = myMin
    if betaInv == 0.0:
        print("bad initialization (0 variance), "
              "setting variance to random number...")
        betaInv = abs(r.uniform(-1, 1, size=1))
    return(betaInv)


def initBetaInvRandom(matD, n_nodes, n_individuals, n_dimensions):
    betaInv = np.sum(matD*1/n_nodes)/(n_individuals*n_dimensions)
    return(betaInv)
